from altex_be.sgrna_designer import BaseEditor
import argparse
import pandas as pd
from pathlib import Path

def parse_base_editors(args: argparse.Namespace) -> dict[str, BaseEditor] | None:
    return {
            args.base_editor_name: BaseEditor(
                base_editor_name=args.base_editor_name,
                pam_sequence=args.base_editor_pam.upper(),
                editing_window_start_in_grna=int(args.base_editor_window_start),
                editing_window_end_in_grna=int(args.base_editor_window_end),
                base_editor_type=args.base_editor_type.lower(),
            )
        }

def parse_gene_file(gene_file: Path) -> set[str] | None:
    if not gene_file:
        return None
    if gene_file.suffix.lower() not in [".txt", ".tsv", ".csv"]:
        raise ValueError("Unsupported file extension for gene file. Use .txt, .tsv, or .csv")
    with open(gene_file, "r") as f:
        interest_genes = {line.strip() for line in f if line.strip()} #空の行は if line.strip がFalseになるので除外できる
    return interest_genes

def show_base_editors_info(base_editors: dict[str, BaseEditor]):
    if base_editors is None:
        raise ValueError("No base editors available to display.")

    for base_editor in base_editors.values():
        print(f"  - {base_editor.base_editor_name} (Type: {base_editor.base_editor_type}, PAM: {base_editor.pam_sequence}, "
            f"Window: {base_editor.editing_window_start_in_grna}-{base_editor.editing_window_end_in_grna})")
        

def get_base_editors_from_args(args: argparse.Namespace) -> dict[str, BaseEditor] | None:
    """
    base editorの情報を含むファイルのパスを示す引数を受け取り、BaseEditorのリストを返す。
    csvまたはtxt, tsv形式のファイルをサポートする
    """
    expected_columns = [
    "base_editor_name",
    "pam_sequence",
    "editing_window_start",
    "editing_window_end",
    "base_editor_type"
    ]
    if not args.base_editor_files:
        return None
    else:
        ext = Path(args.base_editor_files).suffix.lower()

    if ext not in [".csv", ".tsv", ".txt"]:
        raise ValueError("Unsupported file extension for base editor file. Use .csv, .tsv, or .txt")
    if ext in [".csv"]:
        be_df = pd.read_csv(args.base_editor_files, header=0)
    elif ext in [".tsv", ".txt"]:
        be_df = pd.read_csv(args.base_editor_files, sep=None, engine="python", header=0)

    # 列名が期待通りかチェック、違うならエラーを投げる
    if set(be_df.columns) != set(expected_columns):
        raise ValueError(
            f"Base editor file columns are invalid. "
            f"Expected columns: {expected_columns}, but got: {list(be_df.columns)}"
        )
    else:
        return {
            row["base_editor_name"]: BaseEditor(
                base_editor_name=row["base_editor_name"],
                pam_sequence=row["pam_sequence"],
                editing_window_start_in_grna=int(row["editing_window_start"]),
                editing_window_end_in_grna=int(row["editing_window_end"]),
                base_editor_type=row["base_editor_type"],
            )
            for _, row in be_df.iterrows()
        }

def check_input_output_directories(refflat_path: Path, fasta_path: Path, output_directory: Path):
    if not refflat_path.is_file():
        raise FileNotFoundError(f"The provided refFlat file '{refflat_path}' does not exist.")
    if not fasta_path.is_file():
        raise FileNotFoundError(f"The provided FASTA file '{fasta_path}' does not exist.")
    if not output_directory.is_dir():
        raise NotADirectoryError(f"The provided output directory '{output_directory}' does not exist.")

def load_supported_assemblies() -> list[str]:
    """
    パッケージ内のcrispr_direct_supported_assemblies.txtを読み込み、アセンブリ名リストを返す
    """
    # このファイルと同じディレクトリにあるtxtを参照
    txt_path = Path(__file__).parent / "crispr_direct_supported_assemblies.txt"
    with open(txt_path, encoding="utf-8") as f:
        supported_assemblies = {line.strip() for line in f if line.strip() and not line.startswith("#")}
    return supported_assemblies

def is_supported_assembly_name_in_crispr_direct(assembly_name:str) -> bool:
    """
    Purpose: ユーザーが入力したアセンブリ名がCRISPRdirectでサポートされているかを確認する。
    Parameter: 
            assembly_name (str): ユーザーが入力したアセンブリ名
            supported_assemblies (list[str]): CRISPRdirectでサポートされているアセンブリ名のリスト 長いので外部txtとして保存
    return : bool
    """
    supported_assemblies = load_supported_assemblies()
    if assembly_name not in supported_assemblies:
        return False
    return True

def split_df_by_column_chunks(df: pd.DataFrame, chunk_sizes=[12, 6, 6]) -> list[pd.DataFrame]:
    """
    DataFrameをchunk_sizesで指定したカラム数ごとに分割し、各DataFrameをリストで返す。
    """
    columns = df.columns.tolist()
    idx = 0
    df = df.head()
    dfs = []
    for size in chunk_sizes:
        cols = columns[idx:idx+size]
        if not cols:
            break
        dfs.append(df[cols].copy())
        idx += size
    # 残りのカラムも追加
    if idx < len(columns):
        cols = columns[idx:]
        dfs.append(df[cols].copy())
    return dfs


from __future__ import (
    annotations,  # python 3.8以下の型ヒントの頭文字は大文字でないといけない
)

import pandas as pd


def classify_splicing_event(
    target_exon: tuple[int:int],
    all_transcripts: list[list[tuple[int, int]]],
) -> str:
    """
    Purpose:
        タプル (start, end)の形式で与えられたexonが、ある遺伝子の全てのトランスクリプトに含まれるexonの(start, end)のタプルのリストに対して、
        どのようなsplicing eventに該当するかを判定する。
    Parameters:
        target_exon: タプル (start, end)
        all_transcripts: ある遺伝子の全てのトランスクリプトの (start, end)のタプルのリスト (次の関数で遺伝子ごとにグループ化してこの関数にinputする)
    Returns:
        exon_type: str
    """
    target_start, target_end = target_exon

    exact_match = 0
    has_start_match_only = False
    has_start_match_only_and_exist_later_end = False
    has_end_match_only = False
    has_end_match_only_and_exist_earlier_start = False
    has_overlap_without_startend_match = False

    for transcript in all_transcripts:
        if target_exon in transcript:
            exact_match += 1
            continue
        for exon in transcript:
            if (
                exon[0] == target_start and exon[1] != target_end
            ):  # ex[0]は比較対象のexonのstart,ex[1]は比較対象のexonのend
                has_start_match_only = (
                    True  # start だけ他のエキソンと一致し、 end は一致しない
                )
                if exon[1] > target_end: #startが一致するが、endが他のエキソンのendよりも小さい場合
                    has_start_match_only_and_exist_later_end = True
            elif exon[1] == target_end and exon[0] != target_start:
                has_end_match_only = (
                    True  # endだけ他のエキソンと一致し、startは一致しない場合
                )
                if exon[0] < target_start: # endが一致するが、startが他のエキソンのstartよりも大きい場合
                    has_end_match_only_and_exist_earlier_start = True
            elif (exon[0] < target_end and exon[1] > target_start) and (exon[0] != target_start or exon[1] != target_end):
                has_overlap_without_startend_match = True  # start, endどちらも他のエキソンと一致しないが、他のエキソンと1塩基以上の重複が生じている

    total = len(all_transcripts)

    if exact_match == total:
        return (
            "constitutive"  # そもそもsplicing variantがない場合は全てconstitutiveとなる
        )
    elif (
        exact_match > 1
        and exact_match != total
        and not has_start_match_only
        and not has_end_match_only
        and not has_overlap_without_startend_match
    ):
        return "skipped"  # 2つ以上のトランスクリプトに存在するが、全ての転写物には存在しないエキソン
    if has_start_match_only and not has_end_match_only and has_start_match_only_and_exist_later_end:
        return "a5ss-short" # ほかのエキソンとstartが一致するが、endはstartが一致する他のエキソンのendよりも小さい場合 例: [100,200],[100,300]の場合、[100,200]がa5ss-short, [100,300]がa5ss-long
    if has_start_match_only and not has_end_match_only and not has_start_match_only_and_exist_later_end:
        return "a5ss-long" # ほかのエキソンとstartが一致するが、endはstartが一致するエキソンの中で一番長い
    if has_end_match_only and not has_start_match_only and has_end_match_only_and_exist_earlier_start:
        return "a3ss-short" # ほかのエキソンとendが一致するが、startはendが一致する他のエキソンのstartよりも大きい場合 例: [100,200],[150,200]の場合、[100,200]がa3ss-long, [150,200]がa3ss-short
    if has_end_match_only and not has_start_match_only and not has_end_match_only_and_exist_earlier_start:
        return "a3ss-long"
    if has_start_match_only and has_end_match_only:
        return "intron_retention"
    if has_overlap_without_startend_match:
        return "overlap"
    if exact_match == 1 and exact_match != total:
        return "unique"  # 他のトランスクリプトには全く見られないエキソン
    else:
        return "other"


def classify_splicing_events_per_gene(refflat: pd.DataFrame) -> pd.DataFrame:
    result = []

    """
    classify_exon_types()は入力が(start, end)のタプルであることを前提としているので、
    refflatのデータフレームをそのまま入力しても正しく動作しない。
    refflatの各行に対して、exons列を(start, end)のタプルのリストに変換する。
    そしてclassify_row_exons()関数を適用して、各行のexons列に対して分類を行う。
    その結果を新しい列"exontype"に追加する。
    そのリストをpd.concat()で結合して、最終的なDataFrameを返す。
    """

    def classify_row_exons(row, gene_group):
        all_transcripts = gene_group["exons"].tolist()
        return [classify_splicing_event(exon, all_transcripts) for exon in row["exons"]]

    for gene, group in refflat.groupby("geneName"):
        group = group.copy()
        group["exontype"] = group.apply(
            lambda row: classify_row_exons(row, group), axis=1
        )
        result.append(group)

    return pd.concat(result, ignore_index=True)


def flip_a3ss_a5ss_on_minus_strand(classified_refflat: pd.DataFrame) -> pd.DataFrame:
    """
    purpose:
        strandが-の遺伝子では、遺伝子の方向性が逆転するため、転写物のa3ssとa5ssが入れ替わるが、
        classify_exons_per_gene()では方向性を考慮していない。したがってstrandが-のa3ssとa5ssを入れ替える
    Parameter:
        classified_refflat: exontype列とstrand列を持つpd.DataFrame
    Return
        pd.DataFrame
    """

    flip_dict = {"a3ss-short": "a5ss-short", "a5ss-short": "a3ss-short", "a3ss-long": "a5ss-long", "a5ss-long": "a3ss-long"}
    mask = classified_refflat["strand"] == "-"

    # apply: リスト内の a3ss/a5ss を flip_dict で置換
    classified_refflat.loc[mask, "exontype"] = classified_refflat.loc[
        mask, "exontype"
    ].apply(lambda types: [flip_dict.get(t, t) for t in types])

    return classified_refflat

def classify_splicing_events(refflat: pd.DataFrame) -> pd.DataFrame:
    """
    このモジュールのwrap関数
    """
    classified_refflat = classify_splicing_events_per_gene(refflat)
    classified_refflat = flip_a3ss_a5ss_on_minus_strand(classified_refflat)
    return classified_refflat



import pandas as pd
import pybedtools



def annotate_sequence_to_bed(bed: pd.DataFrame, fasta_path: str) -> pd.DataFrame:
    """
    Purpose:
        BED形式のデータに指定される遺伝子座位を参照して、塩基配列をFASTAから取得し、bedに塩基配列を追加する
    Parameters:
        bed: pd.DataFrame, BED形式のデータ(内部でpybedtoolsを使用するため、pybedtools.BedToolに変換可能な形式である必要がある,
            例: ['chrom', 'chromStart', 'chromEnd', 'name', 'score', 'strand']の列を持つ)
        fasta_path: str, FASTAファイルのパス
    Returns:
        bed_for_df: pd.DataFrame, 配列アノテーションが追加されたデータフレーム(形式はBED)
    """
    bed_for_df = bed
    bed_for_sequence = pybedtools.BedTool.from_dataframe(bed)
    # FASTAファイルから配列を取得　s = true　で配列のstrandを考慮し、-の時は相補鎖を出力する
    # もちろん、相補鎖も5'-3'の方向に出力される (今後間違えないように注意する)
    fasta_sequences = bed_for_sequence.sequence(fi=fasta_path, s=True, name=True)
    # 得られた配列ををlistにする(最終的に並べ替えていないBEDに付加するので、keyは必要ない)
    sequences = []
    # ヘッダー行で始まっているときはそれまでに溜めたseqをlistに追加し、seqを初期化
    # それ以外の行は塩基配列としてseqに追加
    with open(fasta_sequences.seqfn) as f:
        seq = []
        for line in f:
            if line.startswith(">"):
                if seq:
                    sequences.append("".join(seq))
                    seq = []
            else:
                seq.append(line.strip())  # 改行を除去してseqに追加
        # 以上の処理は次のheaderに到達したときにlistにseqを追加する
        # 最後の配列は次のheaderが存在しないため、for ループを抜けた後にリストに追加する必要がある
        if seq:
            sequences.append("".join(seq))
    # .sequence()メソッドを使用すると、元の構造が保持されないため、bedの別のコピーをdfにする
    bed_for_df.columns = ["chrom", "chromStart", "chromEnd", "name", "score", "strand"]
    # 取得した配列をデータフレームに追加
    bed_for_df["sequence"] = sequences
    return bed_for_df


def join_sequence_to_single_exon_df(
    single_exon_df: pd.DataFrame,
    acceptor_bed_with_sequences: pd.DataFrame,
    donor_bed_with_sequences: pd.DataFrame,
) -> pd.DataFrame:
    """
    Purpose:
        single_exon_dfにbed_for_dfで取得した塩基配列をleft joinして、acceptor_sequenceまたはdonor_sequence列を追加する
    Parameters:
        single_exon_df: pd.DataFrame, アノテーションを与えるデータフレーム
        acceptor_bed_with_sequences: pd.DataFrame, acceptorのBED形式のデータフレームに配列アノテーションが追加されたもの
        donor_bed_with_sequences: pd.DataFrame, donorのBED形式のデータフレーム
    Returns:
        single_exon_df: pd.DataFrame, 配列アノテーションが追加されたデータフレーム
    """

    acceptor_bed_with_sequences = acceptor_bed_with_sequences.rename(
        columns = {"name":"uuid","chromStart": "chromStart_acceptor", "chromEnd": "chromEnd_acceptor"})
    
    donor_bed_with_sequences = donor_bed_with_sequences.rename(
        columns = {"name":"uuid","chromStart": "chromStart_donor", "chromEnd": "chromEnd_donor"})

    for label, bed in [
        ("acceptor", acceptor_bed_with_sequences),
        ("donor", donor_bed_with_sequences),
    ]:
        single_exon_df = single_exon_df.merge(
            bed[["uuid", f"chromStart_{label}", f"chromEnd_{label}", "sequence"]], on="uuid", how="left"
        ).rename(columns={"sequence": f"{label}_exon_intron_boundary_±25bp_sequence"})
    return single_exon_df

def annotate_sequence_to_splice_sites(
    single_exon_df: pd.DataFrame,
    splice_acceptor_single_exon_df: pd.DataFrame,
    splice_donor_single_exon_df: pd.DataFrame,
    fasta_path: str
) -> pd.DataFrame:
    """
    このモジュールの操作をまとめて実行するためのラッパー関数
    """
    acceptor_bed_with_sequences = annotate_sequence_to_bed(
        splice_acceptor_single_exon_df, fasta_path
    )
    donor_bed_with_sequences = annotate_sequence_to_bed(
        splice_donor_single_exon_df, fasta_path
    )
    single_exon_df = join_sequence_to_single_exon_df(single_exon_df, acceptor_bed_with_sequences, donor_bed_with_sequences)
    return single_exon_df



import argparse
import pandas as pd
from pathlib import Path
import logging
import sys
import datetime
from . import (
    cli_setting,
    refflat_preprocessor,
    sequence_annotator,
    splicing_event_classifier,
    target_exon_extractor,
    sgrna_designer,
    output_formatter,
    offtarget_scorer,
    bed_for_ucsc_custom_track_maker,
    logging_config # noqa: F401
)


def main():
    parser = argparse.ArgumentParser(
        description="Altex BE: A CLI tool for processing refFlat files and extracting target exons.",
    )    
    # 明示的に -v/--version を追加
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version="0.1.0",
        help="Show the version of Altex BE",
    )
    # コマンドライン引数を追加
    dir_group = parser.add_argument_group("Input/Output Options")
    dir_group.add_argument(
        "-r", "--refflat-path",
        required=True,
        help="Path of refflat file"
    )
    dir_group.add_argument(
        "-f", "--fasta-path",
        required=True,
        help="Path of FASTA file"
    )
    dir_group.add_argument(
        "-o", "--output-dir",
        required=True,
        help="Directory of the output files"
    )
    gene_group = parser.add_argument_group("Gene Options")
    gene_group.add_argument(
        "--gene-symbols",
        nargs="+",
        help="List of interest gene symbols (space-separated)"
    )
    gene_group.add_argument(
        "--refseq-ids",
        nargs="+",
        help="List of interest gene Refseq IDs (space-separated)"
    )
    gene_group.add_argument(
        "-a", "--assembly-name",
        default=None,
        required=True,
        help="Name of the genome assembly to use"
    )
    gene_group.add_argument(
        "--gene-file",
        default=None,
        required=False,
        help="Path to a file (csv,txt,tsv) containing gene symbols or Refseq IDs (one gene per line)"
    )
    base_editors = parser.add_argument_group("Base Editor Options")
    base_editors.add_argument(
        "-n", "--be-name",
        default=None,
        required=False,
        help="Name of the base editor to optional use",
    )
    base_editors.add_argument(
        "-p", "--be-pam",
        default=None,
        required=False,
        help="PAM sequence for the base editor",
    )
    base_editors.add_argument(
        "-s", "--be-start",
        default=None,
        required=False,
        help="Window start for the base editor (Count from next to PAM)",
    )
    base_editors.add_argument(
        "-e", "--be-end",
        default=None,
        required=False,
        help="Window end for the base editor (Count from next to PAM)",
    )
    base_editors.add_argument(
        "-t", "--be-type",
        default=None,
        required=False,
        help="Choose the type of base editor, this tool supports ABE and CBE",
    )
    base_editors.add_argument(
        "--be-preset",
        default=None,
        required=False,
        help="Preset for the base editor",
    )
    base_editors.add_argument(
        "--be-files",
        default=None,
        required=False,
        help="input the path of csv file or txt file of base editor information",
    )

    args = parser.parse_args()

    refflat_path = Path(args.refflat_path)
    fasta_path = Path(args.fasta_path)
    output_directory = Path(args.output_dir)
    gene_file = Path(args.gene_file) if args.gene_file else None

    cli_setting.check_input_output_directories(refflat_path, fasta_path, output_directory)

    genes_from_file = cli_setting.parse_gene_file(gene_file) if gene_file else set()
    gene_symbols = set(args.gene_symbols) if args.gene_symbols is not None else set()
    refseq_ids = set(args.refseq_ids) if args.refseq_ids is not None else set()
    interest_gene_list = gene_symbols | refseq_ids | genes_from_file
    if not interest_gene_list:
        parser.error("Please provide at least one interest gene symbol or Refseq ID.")

    preset_base_editors = sgrna_designer.make_preset_base_editors()

    # BaseEditorの決定
    base_editors = {}
    if args.be_files:
        base_editors = cli_setting.get_base_editors_from_args(args)
    
    if args.be_preset is not None:
        if args.be_preset not in preset_base_editors:
            parser.error(f"Invalid base editor preset: {args.be_preset}. Available presets are: {list(preset_base_editors.keys())}")
        else:
            base_editor = preset_base_editors[args.be_preset]
            base_editors[base_editor.base_editor_name] = base_editor

    if args.be_name or args.be_pam or args.be_start or args.be_end or args.be_type:
        if not all([args.base_editor_name, args.base_editor_pam, args.base_editor_window_start, args.base_editor_window_end, args.base_editor_type]):
            parser.error(
            "Base editor information is incomplete. Please provide all required parameters."
            )
        base_editors.update(cli_setting.parse_base_editors(args))

    assembly_name = str(args.assembly_name)
    if not cli_setting.is_supported_assembly_name_in_crispr_direct(assembly_name):
        logging.warning(f"your_assembly : {assembly_name} is not supported by CRISPRdirect. please see <https://crispr.dbcls.jp/doc/>")
    
    output_track_name = f"{datetime.datetime.now().strftime('%Y%m%d%H%M')}_{assembly_name}_sgrnas_designed_by_altex-be"

    if not base_editors:
        parser.error("No base editors specified. Please provide at least one base editor.")

    logging.info("Designing sgRNAs for the following base editors:")
    cli_setting.show_base_editors_info(base_editors)

    logging.info(f"Using this FASTA file as reference genome: {fasta_path}")
    logging.info("-" * 50)
    logging.info("loading refFlat file...")
    refflat = pd.read_csv(
            refflat_path,
            sep="\t",
            header=None,
            names=[
                "geneName",
                "name",
                "chrom",
                "strand",
                "txStart",
                "txEnd",
                "cdsStart",
                "cdsEnd",
                "exonCount",
                "exonStarts",
                "exonEnds",
            ],
        )
    
    logging.info("running processing of refFlat file...")
    refflat = refflat.drop_duplicates(subset=["name"], keep=False)
    refflat = refflat_preprocessor.preprocess_refflat(refflat, interest_gene_list)
    if not refflat_preprocessor.validate_filtered_refflat(refflat, interest_gene_list) :
        logging.warning("your interest gene is not targetable. Exiting...")
        sys.exit(0)

    logging.info("-" * 50)

    logging.info("Classifying splicing events...")

    classified_refflat = splicing_event_classifier.classify_splicing_events(refflat)
    del refflat
    logging.info("-" * 50)
    logging.info("Extracting target exons...")
    splice_acceptor_single_exon_df, splice_donor_single_exon_df, exploded_classified_refflat = target_exon_extractor.wrap_extract_target_exon(classified_refflat)
    if splice_acceptor_single_exon_df.empty and splice_donor_single_exon_df.empty:
        logging.warning("No target exons found for all of the given genes, exiting")
        sys.exit(0)
    for gene in interest_gene_list:
        if gene not in exploded_classified_refflat['geneName'].values:
            logging.info(f"No target exons found for the gene: {gene}. Further processing of {gene} will be skipped.")
        else:
            logging.info(f"Target exons found for the gene: {gene}.")
    logging.info("-" * 50)

    logging.info("Annotating sequences to dataframe from genome FASTA...")
    target_exon_df_with_acceptor_and_donor_sequence = sequence_annotator.annotate_sequence_to_splice_sites(
        exploded_classified_refflat, splice_acceptor_single_exon_df, splice_donor_single_exon_df, fasta_path
    )
    del splice_acceptor_single_exon_df, splice_donor_single_exon_df
    
    logging.info("designing sgRNAs...")
    target_exon_df_with_sgrna_dict = sgrna_designer.design_sgrna_for_base_editors_dict(
        target_exon_df=target_exon_df_with_acceptor_and_donor_sequence,
        base_editors=base_editors
    )
    logging.info("-" * 50)
    logging.info("Formatting output...")
    formatted_exploded_sgrna_df = output_formatter.format_output(target_exon_df_with_sgrna_dict, base_editors)
    if formatted_exploded_sgrna_df.empty:
        logging.warning("No sgRNAs could be designed for given genes and Base Editors, Exiting")
        sys.exit(0)
    del target_exon_df_with_acceptor_and_donor_sequence, exploded_classified_refflat
    
    logging.info("Scoring off-targets...")
    exploded_sgrna_with_offtarget_info = offtarget_scorer.score_offtargets(formatted_exploded_sgrna_df, assembly_name, fasta_path=fasta_path)
    logging.info("-" * 50)
    logging.info("Saving results...")
    exploded_sgrna_with_offtarget_info.to_csv(output_directory / f"{output_track_name}_table.csv")
    logging.info(f"Results saved to: {output_directory / f'{output_track_name}_table.csv'}")

    logging.info("Generating UCSC custom track...")
    bed_df = bed_for_ucsc_custom_track_maker.format_sgrna_for_ucsc_custom_track(exploded_sgrna_with_offtarget_info)

    output_path = output_directory / f"{output_track_name}_ucsc_custom_track.bed"
    track_description: str = f"sgRNAs designed by AltEx-BE on {datetime.datetime.now().strftime('%Y%m%d')}"

    with open(output_path, "w") as f:
        track_header = f'track name="{output_track_name}" description="{track_description}" visibility=2 itemRgb="On"\n'
        f.write(track_header)
        bed_df.to_csv(f, sep="\t", header=False, index=False, lineterminator='\n')

    logging.info(f"UCSC custom track file saved to: {output_path}")
    
    logging.info("All AltEx-BE processes completed successfully.")
    logging.info("Printing summary of output:")
    summary_dfs = cli_setting.split_df_by_column_chunks(exploded_sgrna_with_offtarget_info, chunk_sizes=[12, 6, 6])
    for sub_df in summary_dfs:
        print(sub_df)  # indexも表示される
        print("-" * 40)
    return

if __name__ == "__main__":
    main()
